from typing import TYPE_CHECKING

from ezbuild.log import debug, info

if TYPE_CHECKING:
    from pathlib import Path


def create_dir_if_not_exists(path: Path) -> None:
    debug(f"Checking if {path} exists")
    if not path.exists():
        debug(f"Creating {path}")
        path.mkdir()
        info(f"Created {path}")
    else:
        debug(f"{path} already exists")



from enum import Enum, auto

import typer

from ezbuild.python_environment import PythonEnvironment


class LogLevel(Enum):
    DEBUG = auto()
    INFO = auto()
    ERROR = auto()


def debug(message: str) -> None:
    if PythonEnvironment.debug():
        return

    typer.echo(f"[{typer.style(LogLevel.DEBUG.name, fg=typer.colors.CYAN)}] {message}")


def info(message: str) -> None:
    typer.echo(f"[{typer.style(LogLevel.INFO.name, fg=typer.colors.GREEN)}] {message}")


def error(message: str) -> None:
    typer.echo(f"[{typer.style(LogLevel.ERROR.name, fg=typer.colors.RED)}] {message}")


def cc(message: str) -> None:
    typer.echo(f"[{typer.style('CC', fg=typer.colors.CYAN)}] {message}")


def cxx(message: str) -> None:
    typer.echo(f"[{typer.style('CXX', fg=typer.colors.CYAN)}] {message}")


def ccld(message: str) -> None:
    typer.echo(f"[{typer.style('CCLD', fg=typer.colors.CYAN)}] {message}")


def cxxld(message: str) -> None:
    typer.echo(f"[{typer.style('CXXLD', fg=typer.colors.CYAN)}] {message}")


def ar(message: str) -> None:
    typer.echo(f"[{typer.style('AR', fg=typer.colors.CYAN)}] {message}")


def ranlib(message: str) -> None:
    typer.echo(f"[{typer.style('RANLIB', fg=typer.colors.CYAN)}] {message}")



import json
import subprocess
from importlib.metadata import version
from pathlib import Path
from shutil import rmtree
from typing import Annotated

import typer

from ezbuild import CompileCommand, Environment, Language, log, utils

cli: typer.Typer = typer.Typer()


def version_callback(value: bool) -> None:
    if value:
        print(f"ezbuild {version('ezbuild')}")
        raise typer.Exit


@cli.command()
def init(
    name: Annotated[
        str | None, typer.Argument(help="Name of the project to initialize")
    ] = None,
) -> None:
    """Initialize a new project."""
    cwd = Path.cwd()
    if name is None:
        name = cwd.name

    log.info(f'Using "{name}" (name of current directory) as the project name')
    log.info(f'Using "{name}" as name of the executable to build')

    if any(cwd.iterdir()):
        log.error(f'Directory "{name}" is not empty')

    log.debug("Writing build.ezbuild")
    with Path.open(cwd / "build.ezbuild", "w") as f:
        f.write(f"""env = Environment()

{name} = env.Program(
    name='{name}',
    sources=['{name}.c'],
)
""")
        log.info("Wrote build.ezbuild")

    log.debug(f" Writing {name}.c")
    with Path.open(cwd / f"{name}.c", "w") as f:
        f.write("""#include <stdio.h>\n

int main() {
    printf("Hello, World!\\n");
    return 0;
}
""")
        log.info(f"Wrote {name}.c")


@cli.command()
def build(
    name: Annotated[
        str | None, typer.Argument(help="Name of the project to initialize")
    ] = None,
) -> None:
    """Build the project."""
    cwd = Path.cwd()
    build_dir = cwd / "build"
    bin_dir = build_dir / "bin"
    lib_dir = build_dir / "lib"
    compile_commands: list[CompileCommand] = []

    if not (cwd / "build.ezbuild").exists():
        log.error("build.ezbuild does not exist")
        raise typer.Exit

    log.debug("Reading build.ezbuild")

    with Path.open(cwd / "build.ezbuild", "r") as f:
        build_ezbuild = f.read()

    namespace: dict[str, object] = {
        "Environment": Environment,
        "Language": Language,
    }
    log.debug("Executing build.ezbuild")
    exec(build_ezbuild, namespace)

    for var_name, value in namespace.items():
        if isinstance(value, Environment):
            log.info(f"Found environment: {var_name}")

            if (
                len(value.programs) == 0
                and len(value.static_libraries) == 0
                and len(value.shared_libraries) == 0
            ):
                log.error(f"Environment {var_name} has no buid targets")
                raise typer.Exit

            log.debug(f"Found {len(value.programs)} programs in environment {var_name}")
            utils.fs.create_dir_if_not_exists(build_dir)

            for program in value.programs:
                log.info(f"Building program {program.name}")

                if Language.C in program.languages:
                    value.ensure_cc()
                    value.ensure_ccld()

                if Language.CXX in program.languages:
                    value.ensure_cc()
                    value.ensure_cxx()
                    value.ensure_ccld()
                    value.ensure_cxxld()

                int_dir = build_dir / program.name
                utils.fs.create_dir_if_not_exists(int_dir)

                local_compile_commands: list[CompileCommand] = []
                for source in program.sources:
                    source_path = cwd / source
                    _temp = int_dir / source
                    compile_command = CompileCommand()
                    compile_command.directory = str(int_dir)
                    compile_command.file = str(cwd / source)
                    compile_command.output = str(
                        _temp.parent / ((int_dir / source).name + ".o")
                    )

                    if source_path.suffix == ".c":
                        compile_command.command = " ".join(
                            [
                                value["CC"],
                                "-c",
                                "-o",
                                compile_command.output,
                                compile_command.file,
                            ]
                        )
                        log.cc(f"{compile_command.file}")

                    elif source_path.suffix in [".cpp", ".cxx", ".cc"]:
                        compile_command.command = " ".join(
                            [
                                value["CXX"],
                                "-c",
                                "-o",
                                compile_command.output,
                                compile_command.file,
                            ]
                        )
                        log.cxx(f"{compile_command.file}")

                    result = subprocess.run(
                        compile_command.command, shell=True, capture_output=True
                    )
                    if result.returncode != 0:
                        log.error("Compilation failed")
                        log.debug(f"Error: {result.stderr.decode()}")

                        raise typer.Exit

                    local_compile_commands.append(compile_command)

                utils.fs.create_dir_if_not_exists(bin_dir)
                if Language.CXX in program.languages:
                    cmd = " ".join(
                        [
                            value["CXXLD"],
                            "-o",
                            str(bin_dir / program.name),
                            *[
                                local_compile_command.output
                                for local_compile_command in local_compile_commands
                            ],
                        ]
                    )
                    log.cxxld(f"{bin_dir / program.name}")

                else:
                    cmd = " ".join(
                        [
                            value["CCLD"],
                            "-o",
                            str(bin_dir / program.name),
                            *[
                                local_compile_command.output
                                for local_compile_command in local_compile_commands
                            ],
                        ]
                    )
                    log.ccld(f"{bin_dir / program.name}")

                result = subprocess.run(cmd, shell=True, capture_output=True)

                if result.returncode != 0:
                    log.error("Linking failed")
                    log.debug(f"Error: {result.stderr.decode()}")
                    raise typer.Exit

                compile_commands.extend(local_compile_commands)

            for static_library in value.static_libraries:
                log.info(f"Building static library {static_library.name}")

                if Language.C in static_library.languages:
                    value.ensure_cc()

                if Language.CXX in static_library.languages:
                    value.ensure_cc()
                    value.ensure_cxx()

                value.ensure_ar()
                value.ensure_ranlib()

                int_dir = build_dir / static_library.name
                utils.fs.create_dir_if_not_exists(int_dir)

                local_compile_commands: list[CompileCommand] = []
                for source in static_library.sources:
                    source_path = cwd / source
                    _temp = int_dir / source
                    compile_command = CompileCommand()
                    compile_command.directory = str(int_dir)
                    compile_command.file = str(cwd / source)
                    compile_command.output = str(
                        _temp.parent / ((int_dir / source).name + ".o")
                    )
                    if source_path.suffix == ".c":
                        compile_command.command = " ".join(
                            [
                                value["CC"],
                                "-c",
                                "-o",
                                compile_command.output,
                                compile_command.file,
                            ]
                        )
                        log.cc(f"{compile_command.file}")

                    elif source_path.suffix in [".cpp", ".cxx", ".cc"]:
                        compile_command.command = " ".join(
                            [
                                value["CXX"],
                                "-c",
                                "-o",
                                compile_command.output,
                                compile_command.file,
                            ]
                        )
                        log.cxx(f"{compile_command.file}")

                    result = subprocess.run(
                        compile_command.command, shell=True, capture_output=True
                    )
                    if result.returncode != 0:
                        log.error("Compilation failed")
                        log.debug(f"Error: {result.stderr.decode()}")

                        raise typer.Exit

                    local_compile_commands.append(compile_command)

                utils.fs.create_dir_if_not_exists(lib_dir)
                cmd = " ".join(
                    [
                        value["AR"],
                        "-rc",
                        str(lib_dir / f"{static_library.name}.a"),
                        *[
                            local_compile_command.output
                            for local_compile_command in local_compile_commands
                        ],
                    ]
                )

                log.ar(f"{lib_dir / f'{static_library.name}.a'}")
                result = subprocess.run(cmd, shell=True, capture_output=True)

                if result.returncode != 0:
                    log.error("Archiving failed")
                    log.debug(f"Error: {result.stderr.decode()}")
                    raise typer.Exit

                cmd = " ".join(
                    [
                        value["RANLIB"],
                        str(lib_dir / f"{static_library.name}.a"),
                    ]
                )

                log.ranlib(f"{lib_dir / f'{static_library.name}.a'}")
                result = subprocess.run(cmd, shell=True, capture_output=True)

                if result.returncode != 0:
                    log.error("Ranlib failed")
                    log.debug(f"Error: {result.stderr.decode()}")
                    raise typer.Exit

                compile_commands.extend(local_compile_commands)

            for shared_library in value.shared_libraries:
                log.info(f"Building shared library {shared_library.name}")

                if Language.C in shared_library.languages:
                    value.ensure_cc()
                    value.ensure_ccld()

                if Language.CXX in shared_library.languages:
                    value.ensure_cc()
                    value.ensure_cxx()
                    value.ensure_ccld()
                    value.ensure_cxxld()

                int_dir = build_dir / shared_library.name
                utils.fs.create_dir_if_not_exists(int_dir)

                local_compile_commands: list[CompileCommand] = []
                for source in shared_library.sources:
                    source_path = cwd / source
                    _temp = int_dir / source
                    compile_command = CompileCommand()
                    compile_command.directory = str(int_dir)
                    compile_command.file = str(cwd / source)
                    compile_command.output = str(
                        _temp.parent / ((int_dir / source).name + ".o")
                    )

                    if source_path.suffix == ".c":
                        compile_command.command = " ".join(
                            [
                                value["CC"],
                                "-fPIC",
                                "-c",
                                "-o",
                                compile_command.output,
                                compile_command.file,
                            ]
                        )
                        log.cc(f"{compile_command.file}")
                    elif source_path.suffix in [".cpp", ".cxx", ".cc"]:
                        compile_command.command = " ".join(
                            [
                                value["CXX"],
                                "-fPIC",
                                "-c",
                                "-o",
                                compile_command.output,
                                compile_command.file,
                            ]
                        )
                        log.cxx(f"{compile_command.file}")

                    result = subprocess.run(
                        compile_command.command, shell=True, capture_output=True
                    )
                    if result.returncode != 0:
                        log.error("Compilation failed")
                        log.debug(f"Error: {result.stderr.decode()}")

                        raise typer.Exit

                    local_compile_commands.append(compile_command)

                utils.fs.create_dir_if_not_exists(lib_dir)

                if Language.CXX in shared_library.languages:
                    cmd = " ".join(
                        [
                            value["CXXLD"],
                            "-shared",
                            "-o",
                            str(lib_dir / f"{shared_library.name}.so"),
                            *[
                                local_compile_command.output
                                for local_compile_command in local_compile_commands
                            ],
                        ]
                    )
                    log.cxxld(f"{lib_dir / f'{shared_library.name}.so'}")
                else:
                    cmd = " ".join(
                        [
                            value["CCLD"],
                            "-shared",
                            "-o",
                            str(lib_dir / f"{shared_library.name}.so"),
                            *[
                                local_compile_command.output
                                for local_compile_command in local_compile_commands
                            ],
                        ]
                    )
                    log.ccld(f"{lib_dir / f'{shared_library.name}.so'}")

                result = subprocess.run(cmd, shell=True, capture_output=True)

                if result.returncode != 0:
                    log.error("Linking failed")
                    log.debug(f"Error: {result.stderr.decode()}")
                    raise typer.Exit

                compile_commands.extend(local_compile_commands)

    log.debug("Writing compile_commands.json")

    with Path.open(build_dir / "compile_commands.json", "w") as f:
        f.write(json.dumps(compile_commands, indent=2))
        log.info("Wrote compile_commands.json")


@cli.command()
def run(
    name: Annotated[str, typer.Argument(help="Name of the project to initialize")],
) -> None:
    """Run the project."""
    cwd = Path.cwd()
    build_dir = cwd / "build"
    bin_dir = build_dir / "bin"

    if not (bin_dir / name).exists():
        build(name=name)

    cmd = f"{bin_dir / name}"
    log.debug(f"Running {cmd}")
    result = subprocess.run(cmd, shell=True)

    if result.returncode != 0:
        log.error("Running failed")
        log.debug(f"Error: {result.stderr.decode()}")
        raise typer.Exit


@cli.command()
def clean():
    """Clean the project."""
    cwd = Path.cwd()
    build_dir = cwd / "build"
    if not build_dir.exists():
        log.info("Nothing to clean")
        raise typer.Exit(0)

    log.info("Cleaning build directory")
    rmtree(build_dir)
    log.info("Cleaned build directory")


@cli.callback(invoke_without_command=True)
def main(
    ctx: typer.Context,
    version: Annotated[
        bool,
        typer.Option("--version", "-V", callback=version_callback, is_eager=True),
    ] = False,
) -> None:
    """Simple Build System."""
    if ctx.invoked_subcommand is None:
        build()


if __name__ == "__main__":
    cli()



from dataclasses import dataclass, field
from shutil import which
from sys import platform
from typing import TYPE_CHECKING, Any

from typer import Exit

from ezbuild.log import debug, error

if TYPE_CHECKING:
    from ezbuild.language import Language


@dataclass
class Program:
    name: str
    languages: list[Language]
    sources: list[str]


@dataclass
class StaticLibrary:
    name: str
    languages: list[Language]
    sources: list[str]


@dataclass
class SharedLibrary:
    name: str
    languages: list[Language]
    sources: list[str]


@dataclass
class Environment:
    programs: list[Program] = field(default_factory=list)
    static_libraries: list[StaticLibrary] = field(default_factory=list)
    shared_libraries: list[SharedLibrary] = field(default_factory=list)
    _vars: dict[str, object] = field(default_factory=dict)

    def __getitem__(self, key: str) -> Any:
        return self._vars.get(key)

    def __setitem__(self, key: str, value: object) -> None:
        self._vars[key] = value

    def __contains__(self, key: str) -> bool:
        return key in self._vars

    def Program(
        self, name: str, languages: list[Language], sources: list[str]
    ) -> Program:
        program = Program(name=name, languages=languages, sources=sources)
        self.programs.append(program)
        return program

    def StaticLibrary(
        self, name: str, languages: list[Language], sources: list[str]
    ) -> StaticLibrary:
        static_library = StaticLibrary(name=name, languages=languages, sources=sources)
        self.static_libraries.append(static_library)
        return static_library

    def SharedLibrary(
        self, name: str, languages: list[Language], sources: list[str]
    ) -> SharedLibrary:
        shared_library = SharedLibrary(name=name, languages=languages, sources=sources)
        self.shared_libraries.append(shared_library)
        return shared_library

    def ensure_cc(self) -> None:
        if not self["CC"]:
            if platform == "linux":
                debug("CC is not set, using cc")

                cc_path = which("cc")
                if not cc_path:
                    error("cc not found")
                    raise Exit

                self["CC"] = cc_path
            else:
                error("Unsupported platform")
                raise Exit
        else:
            debug("CC is set")

    def ensure_ccld(self) -> None:
        if not self["CCLD"]:
            debug("CCLD is not set, using CC")
            self["CCLD"] = self["CC"]
        else:
            debug("CCLD is set")

    def ensure_cxx(self) -> None:
        if not self["CC"]:
            error("CXX requires CC")
            raise Exit

        if not self["CXX"]:
            if platform == "linux":
                debug("CXX is not set, using c++")

                cxx_path = which("c++")
                if not cxx_path:
                    error("c++ not found")
                    raise Exit

                self["CXX"] = cxx_path
            else:
                error("Unsupported platform")
                raise Exit
        else:
            debug("CXX is set")

    def ensure_cxxld(self) -> None:
        if not self["CCLD"]:
            error("CXXLD requires CCLD")
            raise Exit

        if not self["CXXLD"]:
            debug("CXXLD is not set, using CXX")
            self["CXXLD"] = self["CXX"]
        else:
            debug("CXXLD is set")

    def ensure_ar(self) -> None:
        if not self["AR"]:
            if platform == "linux":
                debug("AR is not set, using ar")

                ar_path = which("ar")
                if not ar_path:
                    error("ar not found")
                    raise Exit

                self["AR"] = ar_path
            else:
                error("Unsupported platform")
                raise Exit
        else:
            debug("AR is set")

    def ensure_ranlib(self) -> None:
        if not self["AR"]:
            error("AR requires AR")
            raise Exit

        if not self["RANLIB"]:
            if platform == "linux":
                debug("RANLIB is not set, using ranlib")

                ranlib_path = which("ranlib")
                if not ranlib_path:
                    error("ranlib not found")
                    raise Exit

                self["RANLIB"] = ranlib_path
            else:
                error("Unsupported platform")
                raise Exit
        else:
            debug("RANLIB is set")



from .compile_command import CompileCommand
from .environment import Environment, Program
from .language import Language
from .python_environment import PythonEnvironment

__version__ = "0.1.0"

__all__ = [
    "CompileCommand",
    "Environment",
    "Language",
    "Program",
    "PythonEnvironment",
]



class CompileCommand(dict[str, str]):
    def __init__(
        self,
        directory: str = "",
        command: str = "",
        file: str = "",
        output: str = "",
    ) -> None:
        super().__init__(directory=directory, command=command, file=file, output=output)

    @property
    def directory(self) -> str:
        return self["directory"]

    @directory.setter
    def directory(self, value: str) -> None:
        self["directory"] = value

    @property
    def command(self) -> str:
        return self["command"]

    @command.setter
    def command(self, value: str) -> None:
        self["command"] = value

    @property
    def file(self) -> str:
        return self["file"]

    @file.setter
    def file(self, value: str) -> None:
        self["file"] = value

    @property
    def output(self) -> str:
        return self["output"]

    @output.setter
    def output(self, value: str) -> None:
        self["output"] = value
